/*************************************************************************
	-----------------------------------------------------
           		Copyright 2002, 
	   University of Massachusetts, Amherst. 
	-----------------------------------------------------

This package was created at the University of Massachusetts, Amherst. 
The University of Massachusetts makes no warranty about the suitability 
of this software for any purpose. It is presented on an "as is" basis.

CONDITIONS:

The use of software for commercial purposes is strictly forbidden
without a written consent from its authors or the University of 
Massachusetts, Amherst.

**************************************************************************/
/**CFile***********************************************************************

  FileName    [eliminate.c]

  PackageName [BDS]

  Synopsis    [BnetNode eliminate program]

  Description [Nodes on a Bnetwork can be removed by the following two approaches.
  1) if the node is functional equivalent (or complementary) to other nodes.
     this type of nodes are removed by building global BDDs.
  2) satisfy a eliminate gain threshold.]

  SeeAlso     []

  Author      [Anda Congguang Yang]

  Copyright   [The file was created in Department of Electrical & Computer
	Engineering, University of Massachusetts Amherst. All source codes
	in BDS package are CONFIDENTIAL before a formal release is
	announced. The author reserve all right of the package.]

******************************************************************************/
#include "eliminate.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int BDS_LocalEliminateCollect ARGS((DdManager*,BnetNetwork*,bdsOptions*,st_table*,int,FILE*));
static int BDS_LocalEliminateExecute ARGS((DdManager*,BnetNetwork*,bdsOptions*,st_table*,int));
static int bdsFreeOneBnetNode ARGS((DdManager*,BnetNode*,int));
static int BDS_EliminateNodeOrNot ARGS((DdManager*,BnetNode*,bdsOptions*,st_table*,st_table*,int*,int,FILE*)); 
static int bdsCalculateFanoutBddSize ARGS((DdManager*,BnetNode*,int));
static int BDS_ComposeIntoFanouts ARGS((DdManager*,BnetNode*,bdsOptions*));
static int BDS_CheckEliminateGain ARGS((DdManager*,BnetNode*,bdsOptions*,int,FILE*));
static int BDS_CheckEliminateGainWithOutReorder ARGS((DdManager*,BnetNode*,bdsOptions*));
static int BDS_CheckEliminateGainWithReorder ARGS((DdManager*,BnetNode*,bdsOptions*,FILE*));
static int bdsCreateDummyVar ARGS((DdManager*,BnetNode*,int**,int**,st_table*));


/**Function********************************************************************

  Synopsis    [Iterative eliminate internal node based on BDD complexity]

  Description [Internal nodes are iteratively eliminated. In each iteration,
  all valid candidates are recorded. Then they are merged into their fanouts.
  The elimination criteria is based on # of BDD nodes and width (in the future).
  The threshold is passed in through option*. Return 1 if successful; 0 otherwise]

  SideEffects [network is modified]

  SeeAlso     []

  Last Date   [3/23/99]

*****************************************************************************/
extern
int 
BDS_EliminateBasedOnLocal(
  DdManager *bddmgr,
  BnetNetwork *net,
  bdsOptions *option,
  int globalReady,
  int reorder,
  FILE *recdfp)
{
  int result, i;
  int removableNodes = 1; /* Force the first iteration */
  st_table *tobeRemovedTbl;

  printf("Eliminating internal nodes ");
  (void) fflush(stdout);

  /* Iterative elimination */
  while (removableNodes) {
      tobeRemovedTbl = st_init_table(st_ptrcmp,st_ptrhash);
      
      /* Collect all eliminatable internal nodes */
      result = BDS_LocalEliminateCollect(bddmgr,net,option,tobeRemovedTbl,reorder,recdfp);
      if (result == 0) return(0);

      /* Return # of nodes to be eliminated */
      removableNodes = st_count(tobeRemovedTbl);

      /* Conduct node elimination, modify the boolean network */
      result = BDS_LocalEliminateExecute(bddmgr,net,option,tobeRemovedTbl,globalReady);
      if (result == 0) return(0);

      st_free_table(tobeRemovedTbl);
  }

  printf("Done\n\n");

  return(1);

} /* end of BDS_EliminateBasedOnLocal */

/**Function********************************************************************

  Synopsis    [Collect all internal nodes which can be eliminated]
  other internal nodes. Global BDDs are used to detect functional equivalent
  node. Return 1 if successful; 0 otherwise]

  Description [Local BDDs are used to carry out elimination process. The BDD on 
  a node is composed into its fanouts. The decision of elimination is based on
  elimination gain threshold. Return 1 if successful; 0 otherwise]

  SideEffects []

  SeeAlso     []

  Last Date   [3/24/99]

*****************************************************************************/
static
int
BDS_LocalEliminateCollect(
  DdManager *bddmgr,
  BnetNetwork *net,
  bdsOptions *option,
  st_table *tobeRemovedTbl,
  int reorder,
  FILE *recdfp)
{
  int result;
  lsGen fanouts;
  BnetNode *node,*fanode;
  st_generator *gen;
  st_table *mergeSinkTbl;  /* nodes table which fanins will merged into */
  int eliminateOrNot;     /* 0: don't eliminate; 1: elimnate */

  mergeSinkTbl = st_init_table(st_ptrcmp,st_ptrhash);
  gen = st_init_gen(net->livenodes);
  while (st_gen(gen, (char **) &node, NULL)) {
      if (node->type == BNET_CONSTANT_NODE || node->type == BNET_INPUT_NODE ||
					node->type == BNET_OUTPUT_NODE) continue;

      result = BDS_EliminateNodeOrNot(bddmgr,node,option,tobeRemovedTbl,mergeSinkTbl,\
                                                         &eliminateOrNot,reorder,recdfp);
      if (result == 0) return(0);

      if (eliminateOrNot) { 
	  /* Store the merge sinks. This will prevent other nodes merged into it */
	  fanouts = lsStart(node->fanouts);
	  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
              if (st_add_direct(mergeSinkTbl, (char *) fanode, (char *) NULL) == ST_OUT_OF_MEM) {
	          printf("Out of memory"); return(0);
              }
	  }
	  lsFinish(fanouts);

	  /* Store to be merged nodes */
          if (st_add_direct(tobeRemovedTbl, (char *) node, (char *) NULL) == ST_OUT_OF_MEM) {
	      printf("Out of memory"); return(0);
          }
      }
  }
  st_free_gen(gen);
  st_free_table(mergeSinkTbl);

  return(1);

} /* end of BDS_LocalEliminateCollect */

/**Function********************************************************************

  Synopsis    [Carry out node elimination]

  Description [Node in the tobeRemovedTbl are composed into their fanouts, then
  the node is freed. Boolean network is updated. Return 1 if successful; 0 otherwise]

  SideEffects [The connectivity and overall structure of the Boolean network is changed]

  SeeAlso     []

  Last Date   [3/24/99]

*****************************************************************************/
static
int
BDS_LocalEliminateExecute(
  DdManager *bddmgr,
  BnetNetwork *net,
  bdsOptions *option,
  st_table *tobeRemovedTbl,
  int globalReady)
{
  st_generator *gen;
  BnetNode *node, *fanode, *fanin_node, *fanout_node, *dummy_node;
  lsGen fanouts, fanins, fanin_node_fanout;
  int nodeLocalId;
  DdNode *nodeLocalBdd, *composedBdd, *fanoutBdd;

  /* eliminate node one by one, update the boolean network */
  gen = st_init_gen(tobeRemovedTbl);
  while (st_gen(gen, (char **) &node, NULL)) {

#ifndef DEBUG
      (void) printf(".");
      (void) fflush(stdout);
#endif

      nodeLocalId = node->localVar;
      nodeLocalBdd = node->localBdd;
      fanouts = lsStart(node->fanouts);
      while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {

	  /* Build composed BDDs if rodelim, transfer BDDs otherwise */
	  if (option->rodelim == TRUE) {
	      fanoutBdd = fanode->localBdd;
	      composedBdd = Cudd_bddCompose(bddmgr, fanoutBdd, nodeLocalBdd, nodeLocalId);
	      if (composedBdd == NULL) return(0);
	      Cudd_Ref(composedBdd);
	      Cudd_RecursiveDeref(bddmgr, fanode->localBdd);
	      fanode->localBdd = composedBdd;
	  }
	  else {
	      Cudd_RecursiveDeref(bddmgr, fanode->localBdd);
	      fanode->localBdd = fanode->tmpBdd;
	      fanode->tmpBdd = NULL;
	  }

	  /* Move fanins of node to fanode */
	  fanins = lsStart(node->fanins);
	  while (lsNext(fanins, (lsGeneric *) &fanin_node, NULL) != LS_NOMORE) {
	      if (lsNewEnd(fanode->fanins, (lsGeneric) fanin_node, NULL) != LS_OK) {
		  printf("Out of memory"); return(0);
	      }
	  }
	  lsFinish(fanins);

	  /* Remove possible duplications */
	  (void) lsSort(fanode->fanins, bdsUniqueBnetNode);
	  (void) lsUniq(fanode->fanins, bdsUniqueBnetNode, NULL);

	  /* Remove node from fanode's fanin list */
	  fanins = lsStart(fanode->fanins);
	  while (lsNext(fanins, (lsGeneric *) &fanin_node, NULL) != LS_NOMORE) {
	      if (fanin_node == node) {
		  (void) lsDelBefore(fanins, (lsGeneric*) &dummy_node);
		  break;
              }
	  }
	  lsFinish(fanins);

	  /* Update the fanouts of fanin_node. Add fanode to fanin_node's fanout list */
	  fanins = lsStart(node->fanins);
	  while (lsNext(fanins, (lsGeneric *) &fanin_node, NULL) != LS_NOMORE) {
	      
	      /* The following line is commented out to fix a bug. When the fanout
	      ** of a PI node is changed, the fanout field of that PI should also
	      ** be updated. The bug was found in the implementation of sharing extraction 
	      ** -cyang 5/24/99
	      if (fanin_node->type == BNET_INPUT_NODE) continue;
	      */

	      if (lsNewEnd(fanin_node->fanouts, (lsGeneric) fanode, NULL) != LS_OK) {
		  printf("Out of memory"); return(0);
	      }
	  }
	  lsFinish(fanins);
      }
      lsFinish(fanouts);

      /* Remove node from fanin_node's fanout list before node is freed */
      fanins = lsStart(node->fanins);
      while (lsNext(fanins, (lsGeneric *) &fanin_node, NULL) != LS_NOMORE) {

          /* The following line is commented out. See explanation above. -cyang 5/24/99
	  if (fanin_node->type == BNET_INPUT_NODE) continue;
	  */

          fanin_node_fanout = lsStart(fanin_node->fanouts);
          while (lsNext(fanin_node_fanout, (lsGeneric *) &fanout_node, NULL) != LS_NOMORE) {
	      if (fanout_node == node) {
	          (void) lsDelBefore(fanin_node_fanout, (lsGeneric*) &dummy_node);
	          break;
	      }
          }
          lsFinish(fanin_node_fanout);

	  /* Remove possible duplications */
	  (void) lsSort(fanin_node->fanouts, bdsUniqueBnetNode);
	  (void) lsUniq(fanin_node->fanouts, bdsUniqueBnetNode, NULL);
      }	 
      lsFinish(fanins);

      /* Since fanode already conatins information of node, node can be freed */
      if (!st_delete(net->livenodes, (char **) &node, (char **)NULL)) {
	  return(0);
      }
      (void) bdsFreeOneBnetNode(bddmgr, node, globalReady);
  }
  st_free_gen(gen);

  return(1);

} /* end of BDS_LocalEliminateExecute */

/**Function********************************************************************

  Synopsis    [Free one BnetNode]

  Description []

  SideEffects []

  SeeAlso     []

  Last Date   [3/24/99]

*****************************************************************************/
static
int
bdsFreeOneBnetNode(
  DdManager *bddmgr, 
  BnetNode *node,
  int globalReady)
{
  int i;
  BnetTabline *line, *nextline;


  /* Free all input names on the ORIGINAL bnetwork */
  for (i = 0; i < node->ninp; i++) {
      FREE(node->inputs[i]);
  }
  FREE(node->inputs);

  /* Free table line */
  line = node->f;
  while (line != NULL) {
      nextline = line->next;
      FREE(line->values);
      FREE(line);
      line = nextline;
  }

  FREE(node->name);
  Cudd_RecursiveDeref(bddmgr,node->localBdd);
  if (globalReady && node->dd != NULL)
      Cudd_RecursiveDeref(bddmgr, node->dd);

  lsDestroy(node->fanins, NULL);
  lsDestroy(node->fanouts, NULL);

  FREE(node);

  return(1);

} /* end of bdsFreeOneBnetNode */

/**Function********************************************************************

  Synopsis    [Check if a node can be eliminated.]

  Description [First, check if node is legal or not. "Legal" means several things,
  1) the node is a sink other nodes are going to merged into; 2) its fanouts
  are nodes which other nodes will merged into. 3) its fanouts are going to
  merged into other nodes. 4) all its fanouts are single-variable PO nodes.
  Second, check the eligibility of eliminating the node. This process is based 
  on "elimination gain threshold". For the mean time, number of BDD nodes are used. 
  The threshold is passed in from option. Return 1 if successful; 0 otherwise.]

  SideEffects []

  SeeAlso     []

  Last Date   [3/24/99]

*****************************************************************************/
static
int
BDS_EliminateNodeOrNot(
  DdManager *bddmgr,
  BnetNode *node,
  bdsOptions *option,
  st_table *tobeRemovedTbl,
  st_table *mergeSinkTbl,
  int *eliminateOrNot, /* Return value. 0 don't eliminate; 1 eliminate. */
  int reorder,
  FILE *recdfp)
{
  lsGen fanouts;
  BnetNode *fanode;
  int egain; /* Elimination gain */
  int trivialPO;

  *eliminateOrNot = 1; /* Set flag to eliminate, legality check will disqualified this */

  /* If the node is a sink into which other node will merge. */
  if (st_is_member(mergeSinkTbl, (char *) node)) {
      *eliminateOrNot = 0;
      return(1);
  }

  /* If its fanouts are falled into above mentioned 2 and 3 categories */
  fanouts = lsStart(node->fanouts);
  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      if (st_is_member(mergeSinkTbl,(char*)fanode) || 
                         st_is_member(tobeRemovedTbl,(char*)fanode)) {
	  *eliminateOrNot = 0;
	  break;
      }
  }
  lsFinish(fanouts);
  if (*eliminateOrNot == 0) return(1);

  /* Check if all its fanouts are single-variable PO nodes. If this is the case,
  ** those single-variable PO nodes are used to hold previously allocated pointers.
  ** They are the interface between internal and external, don't eliminate the node.
  */
  trivialPO = 0;
  fanouts = lsStart(node->fanouts);
  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      if (fanode->type == BNET_OUTPUT_NODE
                 && Cudd_IsConstant(Cudd_T(fanode->localBdd))
                 && Cudd_IsConstant(Cudd_E(fanode->localBdd))) trivialPO++;
  }
  lsFinish(fanouts);
  if (trivialPO == lsLength(node->fanouts)) {
      *eliminateOrNot = 0;
      return(1);
  }

  /* Check the elimination gain */ 
  egain = BDS_CheckEliminateGain(bddmgr,node,option,reorder,recdfp);
  if (egain <= option->ethred) {
      *eliminateOrNot = 1;
  }
  else {
      *eliminateOrNot = 0;
      if (option->rodelim != TRUE) {
          /* Clean up intermediate results */
          fanouts = lsStart(node->fanouts);
          while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
	      Cudd_RecursiveDeref(bddmgr,fanode->tmpBdd);
          }
          lsFinish(fanouts);
      }
  }

  return(1);

} /* end of BDS_EliminateNodeOrNot */

/**Function********************************************************************

  Synopsis    [Evaluate the gain of eliminating the node]

  Description [The node is merged into its fanouts. The number of BDD nodes
  on new local BDDs of its fanouts are compared with their old local BDDs'.
  gain is calculated as the current number of BDD node minus the previous
  BDD nodes on the node and all its fanous'. ]

  SideEffects []

  SeeAlso     []

  Last Date   [3/24/99]

*****************************************************************************/
static
int
BDS_CheckEliminateGain(
  DdManager *bddmgr,
  BnetNode *node,
  bdsOptions *option,
  int reorder,
  FILE *recdfp)
{
  int gain;

  if (option->rodelim == TRUE && reorder) 
      gain = BDS_CheckEliminateGainWithReorder(bddmgr, node, option, recdfp);
  else
      gain = BDS_CheckEliminateGainWithOutReorder(bddmgr, node, option);

  return(gain);

} /* end of BDS_CheckEliminateGain */

/**Function********************************************************************

  Synopsis    [Calculate the gain of eliminating node]

  Description [localBdd on node and localBdds on all its fanouts are reordered first.
  Then localBdd on node is composed into all its fanouts. Reorder the composed BDD again.
  Compare the total number of BDD nodes before and after composition. To prevent from
  carrying large number of empty BDD variables along with the reordering, a new bddmgr
  is initialized. All operations mentioned above are carried out in this new bddmgr.
  Unlike BDS_CheckEliminateGainWithReorderOld(), bddPool is used.
  Return gain if successful; exit(2) otherwise]

  SideEffects []

  SeeAlso     []

  Last Date   [3/30/99]

*****************************************************************************/
static
int
BDS_CheckEliminateGainWithReorder(
  DdManager *bddmgr,
  BnetNode *node,
  bdsOptions *option,
  FILE *recdfp)
{
  int result, numFanouts, i;
  int *real2dummy, *dummy2realId, dummyIndex, realIndex, numDummyVar, nodeDummyIndex, *jointSupport;
  DdManager *microbddmgr;
  DdNode *dummyVar, *nodeDummyBdd, **fanoutDummyBdds, **composedBddArray;
  DdNode *nodeLocalBdd, **fanoutBddArray, *jointSupportBdd, *scan;
  lsGen fanouts, fanins;
  BnetNode *fanode;
  int numOnNode, numOnFanoutsOld, numOnFanoutsNew;
  bddPool *nodeBddPool, **fanoutBddPool;
  int gain;

  /* Figure out how many dummyVar may need. Since the supports on different BDDs may
  ** overlap, the union of all supports should be found. 
  */
  nodeLocalBdd = node->localBdd;
  if (nodeLocalBdd == NULL) return(0);
  Cudd_Ref(nodeLocalBdd);

  numFanouts = lsLength(node->fanouts);
  fanoutBddArray = ALLOC(DdNode *, numFanouts + 1);
  fanoutDummyBdds = ALLOC(DdNode *, numFanouts + 1);
  fanoutBddPool = ALLOC(bddPool *, numFanouts);
  composedBddArray = ALLOC(DdNode *, numFanouts);
  fanouts = lsStart(node->fanouts);
  i = 0;
  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      fanoutBddArray[i] = fanode->localBdd;
      i++;
  }
  lsFinish(fanouts);
  fanoutBddArray[i] = nodeLocalBdd;

  /* Get the joint support on all BDDs in the order of level */
  jointSupportBdd = Cudd_VectorSupport(bddmgr, fanoutBddArray, numFanouts + 1);
  if (jointSupportBdd == NULL) return(0);
  Cudd_Ref(jointSupportBdd);

  /* Store all BDDs in bddPool locally */
  nodeBddPool = BDS_StoreBddPoolLocal(bddmgr, node->localBdd);
  fanouts = lsStart(node->fanouts);
  i = 0;
  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      fanoutBddPool[i] = BDS_StoreBddPoolLocal(bddmgr, fanode->localBdd);
      i++;
  }
  lsFinish(fanouts);

  /* Find out dummy ~ real correspondency */
  numDummyVar = Cudd_SupportSize(bddmgr, jointSupportBdd);
  real2dummy = ALLOC(int, bddmgr->size);
  jointSupport = ALLOC(int, numDummyVar);

  scan = jointSupportBdd;
  i = 0;
  while (!cuddIsConstant(scan)) {
      jointSupport[i] = Cudd_NodeReadIndex(scan);
      i++;
      scan = cuddT(scan);
  }
  Cudd_RecursiveDeref(bddmgr, jointSupportBdd);

  microbddmgr = startCudd(option,0);
  if (microbddmgr == NULL) fail("2");

#ifndef DEBUG
  Cudd_SetStdout(microbddmgr, recdfp);
#endif

  /* Create dummy variables in new bddmgr */
  for (i = 0; i < numDummyVar; i++) {
      dummyVar = Cudd_bddNewVar(microbddmgr);
      if (dummyVar == NULL) exit(2);
      Cudd_Ref(dummyVar);

      real2dummy[jointSupport[i]] = i;
  }

  /* Start to build dummy BDDs in the microbddmgr. Get BDD node before compose */
  nodeDummyBdd = bdsBuildDDFromBddPoolLocalWithOutCreateDummy(microbddmgr, nodeBddPool, real2dummy);
  if (nodeDummyBdd == NULL) fail("2");
  nodeDummyIndex = real2dummy[node->localVar];
  for (i = 0; i < numFanouts; i++) {
      fanoutDummyBdds[i] = bdsBuildDDFromBddPoolLocalWithOutCreateDummy(microbddmgr, fanoutBddPool[i], real2dummy);
      if (fanoutDummyBdds[i] == NULL) fail("2");
  }
  fanoutDummyBdds[i] = nodeDummyBdd;
  option->reordering = CUDD_REORDER_SIFT;
  BDS_Reorder(microbddmgr, NULL, option);

  numOnFanoutsOld = Cudd_SharingSize(fanoutDummyBdds, numFanouts + 1);

  /* Compose node BDD into all it fanouts */
  fanouts = lsStart(node->fanouts);
  i = 0;
  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      composedBddArray[i] = Cudd_bddCompose(microbddmgr, fanoutDummyBdds[i], nodeDummyBdd, nodeDummyIndex);
      if (composedBddArray[i] == NULL) fail("2");
      Cudd_Ref(composedBddArray[i]);
      Cudd_RecursiveDeref(microbddmgr, fanoutDummyBdds[i]);
      i++;
  }
  lsFinish(fanouts);
  Cudd_RecursiveDeref(microbddmgr, nodeDummyBdd);
  option->reordering = CUDD_REORDER_SIFT;
  BDS_Reorder(microbddmgr, NULL, option);

  numOnFanoutsNew = Cudd_SharingSize(composedBddArray, numFanouts);

  /* Calculate gain */
  gain = numOnFanoutsNew - numOnFanoutsOld;

  FREE(jointSupport);
  FREE(composedBddArray);
  FREE(real2dummy);
  FREE(fanoutDummyBdds);
  FREE(fanoutBddArray);
  FREE(fanoutBddPool);
  Cudd_Quit(microbddmgr);

  return(gain);

} /* end of BDS_CheckEliminateGainWithReorder */

/**Function********************************************************************

  Synopsis    [Calculate the gain of eliminating node]

  Description [localBdd on node and localBdds on all its fanouts are reordered first.
  Then localBdd on node is composed into all its fanouts. Reorder the composed BDD again.
  Compare the total number of BDD nodes before and after composition. To prevent from
  carrying large number of empty BDD variables along with the reordering, a new bddmgr
  is initialized. All operations mentioned above are carried out in this new bddmgr.
  Return gain if successful; exit(2) otherwise]

  SideEffects []

  SeeAlso     []

  Last Date   [3/28/99]

*****************************************************************************/
static
int
BDS_CheckEliminateGainWithReorderOld(
  DdManager *bddmgr,
  BnetNode *node,
  bdsOptions *option)
{
  int result, numFanouts, i;
  int *real2dummy, *dummy2realId, dummyIndex, realIndex, numDummyVar, nodeDummyIndex;
  DdManager *microbddmgr;
  DdNode *dummyVar, *nodeDummyBdd, **fanoutDummyBdds, *composedBdd;
  lsGen fanouts, fanins;
  BnetNode *fanode;
  int numOnNode, numOnFanoutsOld, numOnFanoutsNew;
  SopPool *nodeBddPool, **fanoutBddPool;
  st_table *visited;
  int gain;

  /* Figure out how many dummyVar may need */
  numDummyVar = lsLength(node->fanins);
  fanouts = lsStart(node->fanouts);
  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      numDummyVar += lsLength(fanode->fanins);
  }
  lsFinish(fanouts);

  real2dummy = ALLOC(int, bddmgr->size);
  dummy2realId = ALLOC(int, numDummyVar);
  visited = st_init_table(st_ptrcmp,st_ptrhash);

  /* Store real localBdd on node and all its fanouts */
  nodeBddPool = BDS_DumpSopLocal(bddmgr, node->localBdd);
  if (nodeBddPool == NULL) fail("2");

  numFanouts = lsLength(node->fanouts);
  fanoutBddPool = ALLOC(SopPool *, numFanouts);
  fanoutDummyBdds = ALLOC(DdNode *, numFanouts + 1);
  fanouts = lsStart(node->fanouts);
  i = 0;
  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      fanoutBddPool[i] = BDS_DumpSopLocal(bddmgr, fanode->localBdd);
      if (fanoutBddPool[i] == NULL) fail("2");
      i++;
  }
  lsFinish(fanouts);

  microbddmgr = startCudd(option,0);
  if (microbddmgr == NULL) fail("2");

/*
#ifndef DEBUG
  Cudd_SetStdout(microbddmgr, recdfp);
#endif
*/

  /* Initialize new bdd variables in the new microbddmgr. node first, then its fanouts */
  result = bdsCreateDummyVar(microbddmgr, node, &real2dummy, &dummy2realId, visited);
  if (result == 0) fail("2");

  fanouts = lsStart(node->fanouts);
  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      result = bdsCreateDummyVar(microbddmgr, fanode, &real2dummy, &dummy2realId, visited);
      if (result == 0) fail("2");
  }
  lsFinish(fanouts);

  /* Get the varId to be composed */
  if (!st_lookup(visited, (char *) node, (char **) &dummyVar)) {
      fail("2");
  }
  nodeDummyIndex = Cudd_NodeReadIndex(dummyVar);
  st_free_table(visited);

  /* Start to build dummy BDDs in the microbddmgr. Get BDD node information before compose */
  nodeDummyBdd = BDS_BuildDDFromSopLocal(microbddmgr, nodeBddPool, real2dummy, option);
  if (nodeDummyBdd == NULL) fail("2");
  for (i = 0; i < numFanouts; i++) {
      fanoutDummyBdds[i] = BDS_BuildDDFromSopLocal(microbddmgr, fanoutBddPool[i], real2dummy, option);
      if (fanoutDummyBdds[i] == NULL) fail("2");
  }
  fanoutDummyBdds[i] = nodeDummyBdd;
  option->reordering = CUDD_REORDER_SIFT;
  BDS_Reorder(microbddmgr, NULL, option);

  numOnFanoutsOld = Cudd_SharingSize(fanoutDummyBdds, numFanouts + 1);

  /* Compose node BDD into all it fanouts */
  fanouts = lsStart(node->fanouts);
  i = 0;
  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      composedBdd = Cudd_bddCompose(microbddmgr, fanoutDummyBdds[i], nodeDummyBdd, nodeDummyIndex);
      if (composedBdd == NULL) fail("2");
      Cudd_Ref(composedBdd);
      i++;
  }
  lsFinish(fanouts);
  option->reordering = CUDD_REORDER_SIFT;
  BDS_Reorder(microbddmgr, NULL, option);

  numOnFanoutsNew = Cudd_SharingSize(fanoutDummyBdds, numFanouts);

  /* Constant node 1 is counted twice */
  gain = numOnFanoutsNew - numOnFanoutsOld;

  FREE(real2dummy);
  FREE(dummy2realId);
  FREE(fanoutDummyBdds);
  BDS_FreeSOP(nodeBddPool);
  for (i = 0; i < numFanouts; i++){
      BDS_FreeSOP(fanoutBddPool[i]);
  }
  Cudd_Quit(microbddmgr);

  return(gain);

} /* end of BDS_CheckEliminateGainWithReorderOld */

/**Function********************************************************************

  Synopsis    [Create dummy var]

  Description [Dummy variables are created for all fanins of node. Return 1 if
  successful; 0otherwise]

  SideEffects []

  SeeAlso     []

  Last Date   [3/28/99]

*****************************************************************************/
static
int
bdsCreateDummyVar(
  DdManager *bddmgr, 
  BnetNode *node, 
  int **real2dummy, 
  int **dummy2realId, 
  st_table *visited)
{
  lsGen fanins;
  int dummyIndex, realIndex;
  DdNode *dummyVar;
  BnetNode *fanode;

  fanins = lsStart(node->fanins);
  while (lsNext(fanins, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      if (st_is_member(visited, (char *) fanode)) continue;

      dummyVar = Cudd_bddNewVar(bddmgr);
      if (dummyVar == NULL) return(0);
      Cudd_Ref(dummyVar);
      
      dummyIndex = Cudd_NodeReadIndex(dummyVar);
      realIndex = fanode->localVar;

      (*real2dummy)[realIndex] = dummyIndex;
      (*dummy2realId)[dummyIndex] = realIndex;

      if (st_add_direct(visited, (char *) fanode, (char *) dummyVar) == ST_OUT_OF_MEM) {
          printf("Out of memory");
          return(0);
      }
  }
  lsFinish(fanins);

  return(1);

} /* end of bdsCreateDummyVar */

/**Function********************************************************************

  Synopsis    [Calculate the gain of eliminating node]

  Description [localBdd on node is composed into all its fanouts. Number of BDD
  nodes are compared. Return gain if successful; exit(2) otherwise]

  SideEffects []

  SeeAlso     []

  Last Date   [3/28/99]

*****************************************************************************/
static
int
BDS_CheckEliminateGainWithOutReorder(
  DdManager *bddmgr,
  BnetNode *node,
  bdsOptions *option)
{
  int result, i;
  int numOnFanoutsOld, numOnFanoutsNew, numFanout;
  lsGen fanouts;
  BnetNode *fanode;
  DdNode *nodeBdd, **fanoutBddArray;
  int gain;

  nodeBdd = node->localBdd;
  /* Build BDD array */
  numFanout = lsLength(node->fanouts);
  fanoutBddArray = ALLOC(DdNode *, numFanout + 1);

  i = 0;
  fanouts = lsStart(node->fanouts);
  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      fanoutBddArray[i] = fanode->localBdd;
      i++;
  }
  lsFinish(fanouts);
  fanoutBddArray[i] = nodeBdd;

  numOnFanoutsOld = Cudd_SharingSize(fanoutBddArray, numFanout + 1);

  /* Compose localBdd on node into its fanouts */
  result = BDS_ComposeIntoFanouts(bddmgr,node,option);
  if (result == 0) {
      printf("Bdd composition failed in BDS_CheckEliminateGain!");
      exit(2);
  }

  fanouts = lsStart(node->fanouts);
  i = 0;
  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      fanoutBddArray[i] = fanode->tmpBdd;
      i++;
  }
  lsFinish(fanouts);

  numOnFanoutsNew = Cudd_SharingSize(fanoutBddArray, numFanout);

  /* Constant node 1 is counted twice */
  gain = numOnFanoutsNew - numOnFanoutsOld;

  FREE(fanoutBddArray);
  return(gain);

} /* end of BDS_CheckEliminateGainWithoutReorder */

/**Function********************************************************************

  Synopsis    [Compose the BDD on the node into its fanouts]

  Description [The BDD on this node is composed into its fanouts, one by one.
  Return 1 if successful; 0 otherwise]

  SideEffects [tmpBdd field on the fanout nodes is used to store results.]

  SeeAlso     []

  Last Date   [3/24/99]

*****************************************************************************/
static
int
BDS_ComposeIntoFanouts(
  DdManager *bddmgr,
  BnetNode *node,
  bdsOptions *option)
{
  int retVal = 1;
  BnetNode *fanode;
  lsGen fanouts;
  int localId;       /* index of local variable on the node */
  DdNode *localBdd;  /* local BDD on the node */
  DdNode *fanoutBdd; /* local BDD on the fanout node */
  DdNode *composedBdd;

  localId = node->localVar;
  localBdd = node->localBdd;

  fanouts = lsStart(node->fanouts);
  while (lsNext(fanouts, (lsGeneric *) &fanode, NULL) != LS_NOMORE) {
      fanoutBdd = fanode->localBdd;
      composedBdd = Cudd_bddCompose(bddmgr, fanoutBdd, localBdd, localId);
      if (composedBdd == NULL) {
	  retVal = 0;
	  break;
      }
      else { /* Put the composedBdd into tmpBdd on the fanode */
	  Cudd_Ref(composedBdd);
	  fanode->tmpBdd = composedBdd;
      }
  }
  lsFinish(fanouts);

  return(retVal);

} /* end of BDS_ComposeIntoFanouts */

/**Function********************************************************************

  Synopsis    [Eliminate all internal nodes which are functional duplication of
  other internal nodes. Global BDDs are used to detect functional equivalent
  node. Return 1 if successful; 0 otherwise]

  Description []

  SideEffects []

  SeeAlso     []

  Last Date   [3/23/99]

*****************************************************************************/
extern
int 
BDS_EliminateBasedOnGlobal(
  DdManager *bddmgr,
  BnetNetwork *net,
  bdsOptions *option)
{

  return(1);

} /* end of BDS_EliminateBasedOnGlobal */

/**Function********************************************************************

  Synopsis    [Free Boolean network]

  Description [The Boolean network is different from the original one becasue of
  the eliminate process. Sharing extraction engine may create new Boolean nodes also.]

  SideEffects []

  SeeAlso     []

  Last Date   [8/6/99]

*****************************************************************************/
extern
void
BDS_FreeNetworkAfterEliminate(
  BnetNetwork *net)
{
  int i;
  BnetNode *node;
  BnetTabline *line, *nextline;
  st_generator *gen;

  gen = st_init_gen(net->livenodes);
  while (st_gen(gen, (char **) &node, NULL)) {

      /* Free all input names on the original node */
      for (i = 0; i < node->ninp; i++) {
          FREE(node->inputs[i]);
      }
      if (node->inputs != NULL) FREE(node->inputs);

      /* Free table line */
      if (node->f != NULL) {
          line = node->f;
          while (line != NULL) {
              nextline = line->next;
              FREE(line->values);
              FREE(line);
              line = nextline;
          }
      }

      FREE(node->name);

      if (node->fanins != NULL) lsDestroy(node->fanins, NULL);
      if (node->fanouts != NULL) lsDestroy(node->fanouts, NULL);

      FREE(node);
  }
  st_free_gen(gen);

  return;

} /* end of BDS_FreeNetworkAfterEliminate */




